from djenga.encryption.helpers import _as_bytes
from djenga.encryption.helpers import b64_str
from djenga.encryption.helpers import from_b64_str
from djenga.encryption.helpers import _get_client
from djenga.encryption.helpers import _prefix_alias


def encrypt_bytes(
        plain_text: bytes
        , alias: str
        , region: str=None
        , profile: str=None) -> bytes:
    client = _get_client(region, profile)
    alias = _prefix_alias(alias)
    data = client.encrypt(KeyId=alias, Plaintext=plain_text)
    return data['CiphertextBlob']


def decrypt_bytes(
        cipher_text: bytes
        , region: str=None
        , profile: str=None) -> bytes:
    client = _get_client(region, profile)
    data = client.decrypt(CiphertextBlob=cipher_text)
    return data['Plaintext']


def encrypt(plain_text, alias, region: str=None, profile: str=None) -> str:
    plain_text = _as_bytes(plain_text)
    data = encrypt_bytes(plain_text, alias, region, profile)
    return b64_str(data)


def decrypt(cipher_text: str, region: str=None, profile: str=None):
    cipher_text = from_b64_str(cipher_text)
    data = decrypt_bytes(cipher_text, region, profile)
    return data.decode('utf-8')



import os
from djenga.test import IntegrationTest
from djenga.core import ConfigBunch


class EnvTest(IntegrationTest):
    def test_01(self):
        config = ConfigBunch(
            'env_test_01.yml',
            'env_test_02.yml',
        )
        self.assert_equal(config.dbs.default.username, 'user')
        self.assert_equal(config.dbs.default.host, 'localhost')
        self.assert_equal(config.dbs.default.port, 3306)
        self.assert_equal(config.dbs.default.password, 'my secret value')
        os.environ['DBS_DEFAULT_PASSWORD'] = 'overridden secret value'
        env = config.env()
        self.assert_equal(env('DBS_DEFAULT_PASSWORD'), 'overridden secret value')
        self.assert_equal(env('DBS_DEFAULT_HOST'), 'localhost')
        value = config.get('dbs.default.not_there', 'default')
        self.assert_equal(value, 'default')
        value = config.setdefault('dbs.default.not_there', 'new value')
        self.assert_equal(value, 'new value')
        config['dbs.default.second'] = 'old value'
        self.assert_equal(config.dbs.default.second, 'old value')
        value = config.setdefault('dbs.default.second', 'new value')
        self.assert_equal(value, 'old value')



from typing import Iterable


__all__ = [
    'chunkify',
    'chunkify_iterable',
]


def chunkify(rg, n_max=1000):
    """
    generates sublists of at most n_max
    elements from the provided array
    :param rg: list
    :param n_max: the maximum sublist length
    :return:
    """
    m, n_len = 0, len(rg)
    while m < n_len:
        l, m = m, m + n_max
        yield rg[l:m]


def chunkify_iterable(what: Iterable, n_max: int = 1000):
    """
    generates sublists of at most n_max
    elements from the provided iterable
    """
    n_max = n_max or 1000
    chunk = []
    for x in what:
        chunk.append(x)
        if len(chunk) == n_max:
            yield chunk
            chunk = []
    if chunk:
        yield chunk




import requests
import logging


__all__ = [
    'get_ngrok_url',
]
logger = logging.getLogger(__name__)


def get_ngrok_url():
    """
    Provides a handy helper to get the ngrok public url
    of our dev environment (during debug/testing).
    ngrok has an API that allows us to get our public
    url.  This helper function uses that api call to
    find our public URL and returns it (if ngrok is running).
    Returns None if the function cannot connect to ngrok
    :return: str
    """
    url = None
    response = None
    try:
        response = requests.get('http://127.0.0.1:4040/api/tunnels')
        data = response.json()
        url = data['tunnels'][0]['public_url']
        url = url.replace('http:', 'https:')
        logger.info('NGROK URL = %s', url)
    except:  # noqa: E722
        try:
            if response:
                content = response.content.decode('utf8')
                x = content.find('https')
                y = content.find('\\"', x + 1)
                url = content[x:y]
        except:  # noqa: E722
            pass
    return url



"""
MySQL database backend for Django.

Requires mysqlclient: https://pypi.org/project/mysqlclient/
"""
from django.db.backends.mysql.base import CLIENT, FIELD_TYPE                # isort:skip
from django.db.backends.mysql.base import conversions                      # isort:skip

# Some of these import MySQLdb, so import them after checking if it's installed.
from .client import DatabaseClient                          # isort:skip
from django.db.backends.mysql.creation import DatabaseCreation                      # isort:skip
from django.db.backends.mysql.features import DatabaseFeatures                      # isort:skip
from django.db.backends.mysql.introspection import DatabaseIntrospection            # isort:skip
from django.db.backends.mysql.operations import DatabaseOperations                  # isort:skip
from django.db.backends.mysql.schema import DatabaseSchemaEditor                    # isort:skip
from django.db.backends.mysql.validation import DatabaseValidation                  # isort:skip
from django.db.backends.mysql.base import DatabaseWrapper as ParentDatabaseWrapper
from django.db.backends.mysql.base import version
from django.db.backends.mysql.base import CursorWrapper


class DatabaseWrapper(ParentDatabaseWrapper):
    client_class = DatabaseClient
    def get_connection_params(self):
        from djenga.core import LazySecret
        kwargs = super().get_connection_params()
        password = kwargs.get('passwd')
        if isinstance(password, LazySecret):
            kwargs['passwd'] = password.get()
        return kwargs



from typing import List, Dict
from celery.result import AsyncResult


__all__ = [
    'AsyncDetailedResult',
]


class AsyncDetailedResult(AsyncResult):
    def __init__(self, id, backend=None,
                 task_name=None,            # deprecated
                 app=None, parent=None):
        super().__init__(id, backend, task_name, app, parent)

    def details(self) -> List[Dict]:
        """
        Returns a list of dictionaries with the following keys:
          `key`
          `description`
          `details`
          `latest`
          `time`
          `done`
          `error`
        :return:
        """
        meta = self._get_task_meta()
        return meta.get('details') if meta else []



# encoding: utf-8
from __future__ import absolute_import
from djenga.loggers.util import ColorFormatter
from djenga.loggers.util import BriefColorFormatter
from djenga.loggers.util import IgnoreCaching
from djenga.loggers.util import IgnoreDjangoInternals
from djenga.loggers.util import IgnorePisaPdf
from djenga.loggers.util import IgnoreRequests
from djenga.loggers.util import WorldWritableRotatingFileHandler
from djenga.loggers.filters import *




"""
This module provides functions for rounding numbers as currency.

>>> currency_round_down(1.9999)
Decimal('1.99')
>>> currency_round_up(1.9911)
Decimal('2.00')
>>> currency_round_up4(1.9911)
Decimal('1.9912')
>>> currency_round_half_up(1.9951)
Decimal('2.00')
>>> currency_round_half_up4(1.995555)
Decimal('1.9956')
"""

# encoding: utf-8


from __future__ import unicode_literals
from decimal import Decimal
from decimal import ROUND_FLOOR
from decimal import ROUND_DOWN
from decimal import ROUND_UP
from decimal import ROUND_HALF_UP
from typing import TypeVar


__all__ = [
    'currency_round_down',
    'currency_round_up',
    'currency_round_up4',
    'currency_round_half_up',
    'currency_round_half_up4',
    'round_up',
    'round_down',
    'round_half_up',
    'round_floor',
]


def currency_round_down(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it down to the nearest
    cent.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_FLOOR)
    amount = amount.quantize(Decimal(u'.01'), rounding=ROUND_DOWN)
    return amount


def currency_round_up(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it up to the nearest
    cent.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_FLOOR)
    amount = amount.quantize(Decimal(u'.01'), rounding=ROUND_UP)
    return amount


def currency_round_up4(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it up to the nearest
    one-hundredth of a cent.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_UP)
    return amount


def currency_round_half_up(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it up to the nearest
    cent using half-adjust rounding.  Unlike `currency_round_up`, this
    function will use half-adjust rounding for the cents place after
    floor rounding the ten thousandths place.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_FLOOR)
    amount = amount.quantize(Decimal(u'.01'), rounding=ROUND_HALF_UP)
    return amount


def currency_round_half_up4(amount):
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it up to the nearest
    one-hundredth of a cent using half-adjust rounding.  Unlike
    `currency_round_up4`, this function will use half-adjust rounding
    for the ten thousandths place.
    """
    amount = amount or 0.00
    amount = Decimal(amount)
    amount = amount.quantize(Decimal(u'.0001'), rounding=ROUND_HALF_UP)
    return amount


DecimalLike = TypeVar('DecimalLike', str, float, Decimal)


def q_round(
        amount: DecimalLike,
        places: int=2,
        rounding=ROUND_HALF_UP) -> Decimal:
    """
    Useful helper function that takes a numerical amount, converts
    it to a decimal.Decimal object and rounds it up to the nearest
    one-hundredth of a cent using half-adjust rounding.  Unlike
    `currency_round_up4`, this function will use half-adjust rounding
    for the ten thousandths place.
    """
    amount = amount or Decimal(0.00)
    if not isinstance(amount, Decimal):
        amount = Decimal(amount)
    # This version is slower according to timeit
    # q = Decimal('1') / Decimal(10 ** places)
    # This version, using strings, is actually faster
    if places > 0:
        q = '.%s1' % ('0' * (places - 1),)
    else:
        q = '1'
    q = Decimal(q)
    amount = amount.quantize(q, rounding=rounding)
    return amount


def round_up(amount: DecimalLike, places: int=2) -> Decimal:
    """
    Rounds amount up, to a specified number of places
    :type amount: float | str | decimal.Decimal
    :type places: int
    :rtype: Decimal
    >>> round_up(1.241, 1)
    Decimal('1.3')
    """
    return q_round(amount, places, ROUND_UP)


def round_down(amount, places=2):
    """
    Rounds amount down, to a specified number of places
    :type amount: float | str | decimal.Decimal
    :type places: int
    :rtype: Decimal
    >>> round_down(1.995555, 4)
    Decimal('1.9955')
    >>> round_down('58.12', 0)
    Decimal('58')
    """
    return q_round(amount, places, ROUND_DOWN)


def round_floor(amount: DecimalLike, places: int=2) -> Decimal:
    """
    Floor rounds amount to a specified number of places
    :type amount: float | str | decimal.Decimal
    :type places: int
    :rtype: Decimal
    >>> round_floor(1.995555, 4)
    Decimal('1.9955')
    >>> round_floor('58.12', 0)
    Decimal('58')
    """
    return q_round(amount, places, ROUND_FLOOR)


def round_half_up(amount: DecimalLike, places: int=2) -> Decimal:
    """
    Rounds amount, to a specified number of places, using half-rounding
    :type amount: float | str | decimal.Decimal
    :type places: int
    :rtype: Decimal
    Decimal('1.9956')
    >>> round_half_up(1.9911, 3)
    Decimal('1.991')
    >>> round_half_up(1.995, 2)
    Decimal('2.00')
    """
    return q_round(amount, places, ROUND_HALF_UP)



# encoding: utf-8
from functools import wraps
import logging


__all__ = [
    'update_progress',
    'auto_step',
    'unbound_step',
    'substep',
    'json_logging',
]
logger = logging.getLogger(__name__)


def update_progress(task, progress, *args, **kwargs):
    """
    Adds progress and additional metadata to a celery task.  Since
    celery tasks can also be called as regular ol’ methods, we have
    to handle cases when task.request.id is None (i.e., task.request.id
    is None when a task is called as a regular function instead of with
    `delay` or `apply_async`)
    :param task: celery.Task
    :param progress: format string
    :param args: args for the progress format string
    :param kwargs: additional metadata field/values to set on the task
    """
    if args:
        progress %= args
    request_id = task.request.id
    current = getattr(task.request, 'progress', [])
    current.append(progress)
    if request_id:
        task.backend.mark_as_started(
            request_id,
            progress='\n'.join(current),
            **kwargs)
    setattr(task.request, 'progress', current)
    values = getattr(task.request, 'info', {})
    for key, value in kwargs.items():
        values[key] = value
    setattr(task.request, 'info', values)
    logger.info(progress)


def auto_step(key, description=None,
              start_detail='in progress', end_detail='done'):
    """
    Make sure to apply the `@auto_step` decorator **before** the `@app.task`
    decorator.  e.g.

            @app.task(bind=True, name='my.cool_task',
                      base='djenga.celery.tasks.DetailTask')
            @auto_step(key=1, description='eat cookies')
            def eat_cookies(self):
                pass
    :param key:
    :param description:
    :param start_detail:
    :param end_detail:
    :return:
    """
    def decorator(fn):
        @wraps(fn)
        def decorated(self, *args, **kwargs):
            error = None
            self.start_step(key, description, start_detail)
            try:
                result = fn(self, *args, **kwargs)
                return result
            except Exception as ex:
                error = '%s' % (ex,)
                raise
            finally:
                self.end_step(error, end_detail)
        return decorated
    return decorator


def unbound_step(
        fn, task, key,
        description=None,
        start_detail='in progress',
        end_detail='done'):
    description = description or fn.__name__

    def decorated(*args, **kwargs):
        error = None
        task.start_step(key, description, start_detail)
        try:
            result = fn(*args, **kwargs)
            return result
        except Exception as ex:
            error = '%s' % (ex,)
            raise
        finally:
            task.end_step(error, end_detail)

    return decorated


def substep(task, fn):
    if task:
        task.update_step(fn.__name__)
    return fn


def json_logging():
    """
    Initializes the celery logging so that it uses
    our custom json formatter.
    """
    from celery.signals import after_setup_logger
    from celery.signals import after_setup_task_logger
    from djenga.logging import JsonFormatter
    from djenga.logging import JsonTaskFormatter

    def json_formatter(
            sender, logger: logging.Logger, loglevel, logfile,
            format, colorize, **kw):
        for h in logger.handlers:
            h.formatter = JsonFormatter()

    def json_task_formatter(
            sender, logger: logging.Logger, loglevel, logfile,
            format, colorize, **kw):
        for h in logger.handlers:
            h.formatter = JsonTaskFormatter()

    after_setup_logger.connect(json_formatter)
    after_setup_task_logger.connect(json_task_formatter)


def mark_celery_running():
    """
    marks celery as running by setting the `IS_CELERY_RUNNING`
    flag in the `django.conf` `settings` object.
    """
    from celery.signals import worker_process_init

    def handler(sender, **kwargs):
        from django.conf import settings
        settings.IS_CELERY_RUNNING = True

    worker_process_init.connect(handler)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2018-10-07 19:09
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djenga', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='managementcommand',
            name='last_run',
            field=models.DateTimeField(help_text='The last time this command was run', null=True),
        ),
        migrations.AlterField(
            model_name='managementcommand',
            name='last_success',
            field=models.DateTimeField(help_text='The last time this command was run successfully', null=True),
        ),
        migrations.AlterField(
            model_name='managementcommand',
            name='name',
            field=models.CharField(db_index=True, help_text='The full path of the command', max_length=128),
        ),
        migrations.AlterField(
            model_name='managementcommand',
            name='status',
            field=models.CharField(help_text='The current status of the command', max_length=32),
        ),
    ]



"""
MySQL database backend for Django.

Requires mysqlclient: https://pypi.org/project/mysqlclient/
"""
from django.db.backends.mysql.base import CLIENT, FIELD_TYPE                # isort:skip
from django.db.backends.mysql.base import conversions                      # isort:skip

# Some of these import MySQLdb, so import them after checking if it's installed.
from .client import DatabaseClient                          # isort:skip
from mysql.connector.django.base import DatabaseWrapper as ParentDatabaseWrapper
from mysql.connector.django.base import version


class DatabaseWrapper(ParentDatabaseWrapper):
    client_class = DatabaseClient

    def get_connection_params(self):
        from djenga.core import LazySecret
        kwargs = super().get_connection_params()
        password = kwargs.get('passwd')
        if isinstance(password, LazySecret):
            kwargs['passwd'] = password.get()
        return kwargs
